/**
  *  \file ptrmap.h
  *  \brief Template class ptr_map
  */
#ifndef PCC_V2_PTRMAP_H
#define PCC_V2_PTRMAP_H

#include <map>
#include <iterator>

/** Map of Pointers. Implements ownership semantics for a map of pointer
    values, and attempts to manage it in a type- and exception-safe way.

    The basic assumption is that ptr_map<> owns the values passed to it
    (keys must still manage themselves). It will delete them when the slot
    holding the pointer goes away. You must not delete a pointer obtained
    from ptr_map<>, or place it in another ptr_map<>.

    Otherwise, the interface attempts to be somewhat close to
    std::map<>, restricted to operations that are safe enough to
    maintain ownership semantics. */
template<typename Key, typename T>
class ptr_map {
    ptr_map(const ptr_map&);
    ptr_map& operator=(const ptr_map&);

    /** Content. Implemented by mapping to void*, to share implementation. */
    std::map<Key,void*> data;

    void delete_all();
 public:
    class iterator;
    typedef iterator const_iterator;
    typedef typename std::map<Key,void*>::size_type size_type;
    typedef typename std::map<Key,void*>::difference_type difference_type;
    typedef std::pair<Key, T*> value_type;

    ptr_map();
    ~ptr_map();

    iterator begin() const;
    iterator end() const;

    void insert(const Key& k, T* p);
    void swap(ptr_map& other);
    void swap_elems(iterator a, iterator b);
    T* extract_elem(iterator i);
    size_type size() const;
    size_type max_size() const;
    bool empty() const;
    void clear();
    T* operator[](const Key& k) const;

    iterator find(const Key& k) const;
    size_type count(const Key& k) const;
    void erase(iterator a, iterator e);
    void erase(iterator i);
    size_type erase(const Key& k);
};

/** Bidirectional iterator. Supports
    - post/pre increment/decrement
    - addition and subtraction of difference_type
    - subtraction of another iterator
    - indexing
    - dereferencing
    - comparison */
template<typename Key, typename T>
class ptr_map<Key,T>::iterator
    : public std::iterator<std::bidirectional_iterator_tag,
                           const typename ptr_map<Key,T>::value_type,
                           typename ptr_map<Key,T>::difference_type,
                           const typename ptr_map<Key,T>::value_type*,
                           const typename ptr_map<Key,T>::value_type&>
{
    friend class ptr_map<Key,T>;
    typename std::map<Key,void*>::iterator it;
    struct proxy_type {
        const Key& first;
        T* const second;
        proxy_type(const std::pair<const Key, void*>& orig)
            : first(orig.first), second(static_cast<T*>(orig.second))
            { }
        proxy_type* operator->()
            { return this; }
        const proxy_type* operator->() const
            { return this; }
    };
 public:
    typedef typename std::map<Key,void*>::difference_type difference_type;

    // Constructor
    iterator(typename std::map<Key,void*>::iterator it)
        : it(it) { }

    // Increment and Decrement
    iterator& operator++()
        { ++it; return *this; }
    iterator operator++(int)
        { iterator res(it); ++it; return res; }
    iterator& operator--()
        { --it; return *this; }
    iterator operator--(int)
        { iterator res(it); --it; return res; }

    // Comparison
    bool operator==(iterator rhs) const
        { return it == rhs.it; }
    bool operator!=(iterator rhs) const
        { return it != rhs.it; }

    // Dereferencing
    proxy_type operator*() const
        { return proxy_type(*it); }
    proxy_type operator->() const
        { return proxy_type(*it); }
};


/** Constructor. Makes a blank map. */
template<typename Key, typename T>
ptr_map<Key,T>::ptr_map()
{ }

/** Destructor. Deletes all contained elements. */
template<typename Key, typename T>
ptr_map<Key,T>::~ptr_map()
{
    delete_all();
}

/** Delete all contained elements. */
template<typename Key, typename T>
void
ptr_map<Key,T>::delete_all()
{
    for (typename std::map<Key,void*>::iterator i = data.begin(); i != data.end(); ++i)
        delete static_cast<T*>(i->second);
}

/** Iterator at beginning of map. */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::iterator
ptr_map<Key,T>::begin() const
{
    return iterator(const_cast<std::map<Key,void*>&>(data).begin());
}

/** Iterator at end of map. */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::iterator
ptr_map<Key,T>::end() const
{
    return iterator(const_cast<std::map<Key,void*>&>(data).end());
}

/** Insert new element. Takes ownership of p. If anything throws, p is deleted.
    \todo Note that this is currently not the standard std::map insert() signature.
    \param k Key to create/replace
    \param p Value to store */
template<typename Key, typename T>
void
ptr_map<Key,T>::insert(const Key& k, T* p)
{
    // Not too efficient, but works for the time being
    try {
        delete operator[](k);
        data[k] = p;
    }
    catch (...) {
        delete p;
        throw;
    }
}

/** Swap with another map. */
template<typename Key, typename T>
inline void
ptr_map<Key,T>::swap(ptr_map<Key,T>& other)
{
    data.swap(other.data);
}

/** Swap some elements. The iterators must not be end(). */
template<typename Key, typename T>
inline void
ptr_map<Key,T>::swap_elems(iterator a, iterator b)
{
    void* p = a.it->second;
    a.it->second = b.it->second;
    b.it->second = p;
}

/** Extract an element. Returns the element. The stored element will be erased. */
template<typename Key, typename T>
inline T*
ptr_map<Key,T>::extract_elem(iterator i)
{
    T* p = static_cast<T*>(i.it->second);
    data.erase(i.it);
    return p;
}

/** Get number of elements in this map. */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::size_type
ptr_map<Key,T>::size() const
{
    return data.size();
}

/** Get maximum size (number of elements this map can contain). */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::size_type
ptr_map<Key,T>::max_size() const
{
    return data.max_size();
}

/** Check for emptiness.
    \return true iff map is empty. */
template<typename Key, typename T>
inline bool
ptr_map<Key,T>::empty() const
{
    return data.empty();
}

/** Clear this map. Deletes all contained value pointers. */
template<typename Key, typename T>
inline void
ptr_map<Key,T>::clear()
{
    delete_all();
    data.clear();
}

/** Element access. Note that unlike a std::map, a ptr_map can only be read this way
    (on the plus side, using operator[] to checking for an element will not create it).
    \param k Key to access */
template<typename Key, typename T>
T*
ptr_map<Key,T>::operator[](const Key& k) const
{
    typename std::map<Key,void*>::const_iterator it = data.find(k);
    if (it != data.end())
        return static_cast<T*>(it->second);
    else
        return 0;
}

/** Find element by key. Returns an iterator pointing to an element whose
    key is k. Returns end() if there is no such element. */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::iterator
ptr_map<Key,T>::find(const Key& k) const
{
    return iterator(const_cast<std::map<Key,void*>&>(data).find(k));
}

/** Count elements by key. Returns the number of elements whose key is k.
    For a map, this is either 0 or 1. */
template<typename Key, typename T>
inline typename ptr_map<Key,T>::size_type
ptr_map<Key,T>::count(const Key& k) const
{
    return data.count(k);
}

/** Erase elements given a range. Erases all elements from a
    (includive) up to e (not inclusive). */
template<typename Key, typename T>
inline void
ptr_map<Key,T>::erase(iterator a, iterator e)
{
    while (a != e) {
        iterator m = a;
        ++a;
        this->erase(m);
    }
}

/** Erase element given an iterator. Erases the element pointed-to by
    the iterator. */
template<typename Key, typename T>
void
ptr_map<Key,T>::erase(iterator i)
{
    delete i->second;
    data.erase(i.it);
}

/** Erase element given a key. Erases all elements whose key equals k.
    Returns the number of elements erased. */
template<typename Key, typename T>
typename ptr_map<Key,T>::size_type
ptr_map<Key,T>::erase(const Key& k)
{
    iterator i = this->find(k);
    if (i != this->end()) {
        this->erase(i);
        return 1;
    } else {
        return 0;
    }
}

#endif
